/*
 * Copyright (c) 2015, Nordic Semiconductor
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package no.nordicsemi.android.nrftoolbox.crossdfu;

import android.annotation.SuppressLint;
import android.app.LoaderManager;
import android.bluetooth.BluetoothDevice;
import android.content.CursorLoader;
import android.content.Intent;
import android.content.Loader;
import android.database.Cursor;
import android.graphics.Color;
import android.net.Uri;
import android.os.Bundle;
import androidx.annotation.NonNull;
import androidx.appcompat.app.AlertDialog;
import androidx.appcompat.widget.Toolbar;

import android.os.SystemClock;
import android.provider.MediaStore;
import android.text.TextUtils;
import android.view.View;
import android.webkit.MimeTypeMap;
import android.widget.Button;
import android.widget.Chronometer;
import android.widget.ListView;
import android.widget.ProgressBar;
import android.widget.TextView;
import android.widget.Toast;

import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Objects;
import java.util.UUID;

import no.nordicsemi.android.ble.data.Data;
import no.nordicsemi.android.nrftoolbox.R;
import no.nordicsemi.android.nrftoolbox.dfu.adapter.FileBrowserAppsAdapter;
import no.nordicsemi.android.nrftoolbox.profile.BleProfileActivity;
import no.nordicsemi.android.nrftoolbox.profile.LoggableBleManager;

import static java.lang.StrictMath.min;

// TODO The CrossDfuActivity should be rewritten to use the service approach, like other do.
public class CrossDfuActivity extends BleProfileActivity implements CrossDfuManagerCallbacks,
		LoaderManager.LoaderCallbacks<Cursor> {
	@SuppressWarnings("unused")
	private static final int SELECT_FILE_REQ = 1;
	private static final String EXTRA_URI = "uri";
	private static final char[] hexCode = "0123456789ABCDEF".toCharArray();

	private static final int REQUEST_SIZE = 1;
	private static final int REQ_START_CROSS_DFU = 0X00;
	private static final int REQ_GET_IMAGE_SIZE = 0X01;
	private static final int REQ_GET_IMAGE_DATA = 0x02;
	private static final int BYTES_PER_PACKET = 128;
	private static final int PACKETS_PER_BLOCK = 8;

	private BluetoothDevice selectedDevice;
	private String filePath;
	private Uri fileUri;

	private long imageSize;
	private long imageOffset;
	private int writtenPackets;

	private TextView fileNameView;
	private TextView fileTypeView;
	private TextView fileSizeView;
	private TextView fileStatusView;
	private TextView textPercentage;
	private TextView textUploading;
	private Chronometer timerSendingTime;
	private ProgressBar progressBar;

	private Button selectFileButton, uploadButton, connectButton;

	private CrossDfuManager manager;

	@Override
	protected void onCreateView(final Bundle savedInstanceState) {
		setContentView(R.layout.activity_feature_crossdfu);

		final Toolbar toolbar = findViewById(R.id.toolbar_actionbar);
		setSupportActionBar(toolbar);
		Objects.requireNonNull(getSupportActionBar()).setDisplayHomeAsUpEnabled(true);

		fileNameView = findViewById(R.id.file_name);
		fileTypeView = findViewById(R.id.file_type);
		fileSizeView = findViewById(R.id.file_size);
		fileStatusView = findViewById(R.id.file_status);
		selectFileButton = findViewById(R.id.action_select_file);
		uploadButton = findViewById(R.id.action_upload);
		connectButton = findViewById(R.id.action_connect);
		textPercentage = findViewById(R.id.textviewProgress);
		textUploading = findViewById(R.id.textviewUploading);
		progressBar = findViewById(R.id.progressbar_file);
		timerSendingTime = findViewById(R.id.timerSendingTime);

		imageSize = 0;
	}

	@Override
	protected LoggableBleManager<CrossDfuManagerCallbacks> initializeManager() {
		manager = CrossDfuManager.getCrossDfuManager(getApplicationContext());
		manager.setGattCallbacks(this);
		return manager;
	}

	@Override
	protected void setDefaultUI() {
		boolean dfuFileState;
		dfuFileState = fileStatusView.getText() == getResources().getString(R.string.dfu_file_status_ok);

		if (!dfuFileState) {
			fileNameView.setText("-");
			fileTypeView.setText("-");
			fileSizeView.setText("-");
			fileStatusView.setText(R.string.dfu_file_status_no_file);
		}
	}

	@Override
	protected int getLoggerProfileTitle() {
		return R.string.crossdfu_feature_title;
	}

	@Override
	protected int getDefaultDeviceName() {
		return R.string.crossdfu_default_name;
	}

	@Override
	protected int getAboutTextId() {
		return R.string.crossdfu_about_text;
	}

	@Override
	protected UUID getFilterUUID() {
		return null; /*CrossDfuManager.CROSSDFU_SERVICE_UUID;*/
	}

	@Override
	public void onDeviceReady(@NonNull final BluetoothDevice device) {
		selectedDevice = device;
	}

	@Override
	public void onDeviceDisconnected(@NonNull final BluetoothDevice device) {
		super.onDeviceDisconnected(device);
		selectedDevice = null;
	}

	@Override
	public Loader<Cursor> onCreateLoader(final int id, final Bundle args) {
		final Uri uri = args.getParcelable(EXTRA_URI);
		/*
		 * Some apps, f.e. Google Drive allow to select file that is not on the device. There is no "_data" column handled by that provider. Let's try to obtain
		 * all columns and than check which columns are present.
		 */
		// final String[] projection = new String[] { MediaStore.MediaColumns.DISPLAY_NAME, MediaStore.MediaColumns.SIZE, MediaStore.MediaColumns.DATA };
		return new CursorLoader(this, uri, null /* all columns, instead of projection */, null, null, null);
	}

	@Override
	public void onLoaderReset(final Loader<Cursor> loader) {
		fileNameView.setText(null);
		fileTypeView.setText(null);
		fileSizeView.setText(null);
		filePath = null;
		fileUri = null;
	}

	@Override
	public void onLoadFinished(final Loader<Cursor> loader, final Cursor data) {
		if (data != null && data.moveToNext()) {
			/*
			 * Here we have to check the column indexes by name as we have requested for all. The order may be different.
			 */
			final String fileName = data.getString(data.getColumnIndex(MediaStore.MediaColumns.DISPLAY_NAME)/* 0 DISPLAY_NAME */);
			final int fileSize = data.getInt(data.getColumnIndex(MediaStore.MediaColumns.SIZE) /* 1 SIZE */);
			String path = null;
			final int dataIndex = data.getColumnIndex(MediaStore.MediaColumns.DATA);
			if (dataIndex != -1)
				path = data.getString(dataIndex /* 2 DATA */);
			if (!TextUtils.isEmpty(path))
				this.filePath = path;

			updateFileInfo(fileName, fileSize);
		} else {
			fileNameView.setText(null);
			fileTypeView.setText(null);
			fileSizeView.setText(null);
			filePath = null;
			fileUri = null;
			fileStatusView.setText(R.string.dfu_file_status_error);
		}
	}

	@Override
	protected void onActivityResult(final int requestCode, final int resultCode, final Intent data) {
		super.onActivityResult(requestCode, resultCode, data);
		if (resultCode != RESULT_OK)
			return;

		if (requestCode == SELECT_FILE_REQ) {
			// clear previous data
			filePath = null;

			// and read new one
			final Uri uri = data.getData();
			/*
			 * The URI returned from application may be in 'file' or 'content' schema. 'File' schema allows us to create a File object and read details from if
			 * directly. Data from 'Content' schema must be read by Content Provider. To do that we are using a Loader.
			 */
			if (uri.getScheme().equals("file")) {
				// the direct path to the file has been returned
				final String path = uri.getPath();
				final File file = new File(path);
				filePath = path;

				updateFileInfo(file.getName(), file.length());
			} else if (uri.getScheme().equals("content")) {
				// an Uri has been returned
				fileUri = uri;

				// if application returned Uri for streaming, let's us it. Does it works?
				// FIXME: both Uris works with Google Drive app. Why both? What's the difference? How about other apps like DropBox?
				final Bundle extras = data.getExtras();
				if (extras != null && extras.containsKey(Intent.EXTRA_STREAM))
					fileUri = extras.getParcelable(Intent.EXTRA_STREAM);

				// file name and size must be obtained from Content Provider
				final Bundle bundle = new Bundle();
				bundle.putParcelable(EXTRA_URI, uri);
				getLoaderManager().restartLoader(SELECT_FILE_REQ, bundle, this);
			}
		}
	}

	@SuppressLint("DefaultLocale")
	@Override
	public void onDataReceived(@NonNull BluetoothDevice device, @NonNull Data data) {
		System.out.println(String.format("Received data(%d):", data.size()));
		System.out.println(byteArrayToString(data.getValue(), data.size()));

		if (data.getValue() != null) {
			crossDfuProtoProcess(data.getValue());
		}
	}

	@Override
	public void onSendCompleted(@NonNull BluetoothDevice device) {
		System.out.println("Write request done");

		int percent = (int)(imageOffset * 100 / imageSize);
		runOnUiThread(new Runnable() {
			@Override
			public void run() {
				progressBar.setProgress(percent);
				textPercentage.setText(getString(R.string.dfu_uploading_percentage, percent));
			}
		});

		// The first packet must go after receiving REQ_GET_IMG_DATA
		if (imageOffset == 0) {
			return;
		}

		if (imageSize > imageOffset) {
			sendImagePacket();
		}
		else {
			System.out.println("BLE sending complete");
			textUploading.setText(R.string.crossdfu_status_complete);
			imageOffset = 0;
			fileStatusView.setText(R.string.dfu_file_status_no_file);
			timerSendingTime.stop();
		}
	}

	/**
	 * Updates the file information on UI
	 *
	 * @param fileName file name
	 * @param fileSize file length
	 */
	private void updateFileInfo(final String fileName, final long fileSize) {
		fileNameView.setText(fileName);
		fileSizeView.setText(getString(R.string.dfu_file_size_text, fileSize));

		imageSize = fileSize;
		imageOffset = 0;

		final String extension = "(?i)BIN"; // (?i) =  case insensitive
		final boolean statusOk = MimeTypeMap.getFileExtensionFromUrl(fileName).matches(extension);
		fileTypeView.setText(statusOk ? R.string.crossdfu_file_type : R.string.crossdfu_file_type_unknown);
		fileStatusView.setText(statusOk ? R.string.dfu_file_status_ok : R.string.dfu_file_status_invalid);
		if (!statusOk) {
			fileStatusView.setTextColor(Color.RED);
			fileTypeView.setTextColor(Color.RED);
		}

		textUploading.setText(R.string.crossdfu_status_idle);
		textPercentage.setText(getString(R.string.dfu_uploading_percentage, 0));
	}


	private void openFileChooser() {
		final String MIME_TYPE_BINARY = "application/x-binary";
		String[] mimeTypes = {MIME_TYPE_BINARY};

		final Intent intent = new Intent(Intent.ACTION_GET_CONTENT);
		intent.setType(MIME_TYPE_BINARY);
//		intent.setType("application/x-binary");
		intent.addCategory(Intent.CATEGORY_OPENABLE);
		if (intent.resolveActivity(getPackageManager()) != null) {
			// file browser has been found on the device
			startActivityForResult(intent, SELECT_FILE_REQ);
		} else {
			// there is no any file browser app, let's try to download one
			final View customView = getLayoutInflater().inflate(R.layout.app_file_browser, null);
			final ListView appsList = customView.findViewById(android.R.id.list);
			appsList.setAdapter(new FileBrowserAppsAdapter(this));
			appsList.setChoiceMode(ListView.CHOICE_MODE_SINGLE);
			appsList.setItemChecked(0, true);
			new AlertDialog.Builder(this)
					.setTitle(R.string.dfu_alert_no_filebrowser_title)
					.setView(customView)
					.setNegativeButton(R.string.no, (dialog, which) -> dialog.dismiss())
					.setPositiveButton(R.string.ok, (dialog, which) -> {
						final int pos = appsList.getCheckedItemPosition();
						if (pos >= 0) {
							final String query = getResources().getStringArray(R.array.dfu_app_file_browser_action)[pos];
							final Intent storeIntent = new Intent(Intent.ACTION_VIEW, Uri.parse(query));
							startActivity(storeIntent);
						}
					})
					.show();
		}
	}

	/**
	 * Called when Select File was pressed
	 *
	 * @param view a button that was pressed
	 */
	public void onSelectFileClicked(final View view) {
		openFileChooser();
	}

	private void writeImageData(final byte[] data, final int length) {
		System.out.println(byteArrayToString(data, length));
	}

	@NotNull
	private byte[] buildImagePacket(long offset, long length) {
		int len = (int)length;
		byte data[] = new byte[len + REQUEST_SIZE];

		data[0] = REQ_GET_IMAGE_DATA;

		try {
			InputStream stream = getImageInputStream();
			if (stream == null) {
				System.out.println("Open image file error");
				return data;
			}

			long skipped = stream.skip(offset);
			if (skipped != offset) {
				System.out.println("Seek image offset error");
			}

			if (stream.read(data, REQUEST_SIZE, len) != len) {
				System.out.println("Error at reading image data");
			}
			// Debug
			System.out.println(String.format("%02x %02x %02x %02x", data[0], data[1], data[2], data[3]));
			stream.close();

		} catch (FileNotFoundException e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		} catch (IOException e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}

		return data;
	}

	/**
	 * Callback of UPDATE/CANCEL button on DfuActivity
	 */
	@SuppressLint("DefaultLocale")
	public void onUploadClicked(final View view) {
		boolean fileStatus = (fileStatusView.getText() == getResources().getString(R.string.dfu_file_status_ok));
		boolean deviceStatus = (selectedDevice != null);

		if (fileStatus && deviceStatus && imageOffset == 0) {
			System.out.println("Start to send image file");
			textUploading.setText(R.string.crossdfu_status_sending);

			writtenPackets = 0;
			crossDfuProtoStart();

			timerSendingTime.setBase(SystemClock.elapsedRealtime());
			timerSendingTime.start();
		}
		else {
			popToast("File or device is not ready");
			System.out.println(String.format("imageOffset = %d", imageOffset));
			System.out.println(String.format("file state: %s, device state: %s",
					fileStatus ? "1" : "0",
					deviceStatus ? "1" : "0"));
		}
	}

	InputStream getImageInputStream() throws FileNotFoundException {
		InputStream stream = null;
		if (filePath != null) {
//			stream = openFileInput(filePath);
			stream = new FileInputStream(new File(filePath));
		}
		else if (fileUri != null) {
			stream = getContentResolver().openInputStream(fileUri);
		}

		return stream;
	}


	void crossDfuProtoStart() {
		byte[] req = new byte[1];
		req[0] = REQ_START_CROSS_DFU;

		crossDfuProtoSend(req);
	}

	void crossDfuProtoSend(byte[] data) {
		System.out.println("Write a packet");
		manager.send(data);
	}

	void crossDfuProtoProcess(@NonNull byte[] notification) {
		if (notification.length != REQUEST_SIZE) {
			System.out.println("Unknown request from slave");
			return;
		}
		int req = notification[0];

		if (req == REQ_GET_IMAGE_SIZE) {
			if (imageSize == 0) {
				popToast("No valid image file");
				return;
			}

			byte[] imageSizeBytes = intToByteArray((int) imageSize);

			byte[] response = new byte[5];
			response[0] = REQ_GET_IMAGE_SIZE;
			System.arraycopy(imageSizeBytes, 0, response, 1, imageSizeBytes.length);

			System.out.println("image size response:");
			System.out.println(byteArrayToString(response, response.length));

			crossDfuProtoSend(response);
		} else if (req == REQ_GET_IMAGE_DATA) {
			if (fileUri == null && filePath == null) {
				System.out.println("No image file");
				popToast("No image file");
				return;
			}

			writtenPackets = 0;

			sendImagePacket();
		} else {
			popToast("Unknown request from slave");
		}
	}

	@SuppressLint("DefaultLocale")
	void sendImagePacket() {
		if (imageOffset >= imageSize) {
			return;
		}

		if (writtenPackets == PACKETS_PER_BLOCK) {
			return;
		}

		long len = min(BYTES_PER_PACKET, imageSize - imageOffset);
		System.out.println(String.format("offset = %d, len = %d", imageOffset, len));

		byte[] packet = buildImagePacket(imageOffset, len);
		System.out.println(String.format("offset: %d", imageOffset));

		crossDfuProtoSend(packet);
//		writeImageData(packet, packet.length);

		imageOffset += len;

		writtenPackets += 1;
	}

	@org.jetbrains.annotations.NotNull
	private byte[] intToByteArray(final int i) {
		ByteBuffer buffer = ByteBuffer.allocate(4);
		buffer.order(ByteOrder.LITTLE_ENDIAN);
		buffer.putInt(i);
		return buffer.array();
	}

	public String byteArrayToString(byte[] data, int length) {
		StringBuilder r = new StringBuilder(length * 3);
		for (int i = 0; i < length; i++) {
			r.append(hexCode[(data[i] >> 4) & 0xF]);
			r.append(hexCode[(data[i] & 0xF)]);
			r.append(' ');
		}
		return r.toString();
	}

	private void popToast(final int messageResId) {
		Toast.makeText(this, messageResId, Toast.LENGTH_SHORT).show();
	}

	private void popToast(final String message) {
		Toast.makeText(this, message, Toast.LENGTH_SHORT).show();
	}
}
